# Generated by Django 2.2.19 on 2021-03-15 10:35

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RSAKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('priv_key', models.TextField(help_text='Paste your private RSA Key here.', unique=True, verbose_name='Key')),
            ],
            options={
                'verbose_name': 'RSA Key',
                'verbose_name_plural': 'RSA Keys',
            },
        ),
    ]



import json
import logging
import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist, ValidationError, FieldDoesNotExist
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.signals import post_save, pre_save, pre_delete, m2m_changed
from django.dispatch import receiver
from django.urls import get_resolver
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.decorators import classonlymethod
from guardian.shortcuts import assign_perm
from rest_framework.utils import model_meta
from djangoldp.fields import LDPUrlField
from djangoldp.permissions import DEFAULT_DJANGOLDP_PERMISSIONS, OwnerPermissions, InheritPermissions, ReadOnly

logger = logging.getLogger('djangoldp')

Group._meta.serializer_fields = ['name', 'user_set']
Group._meta.rdf_type = 'foaf:Group'
# Group._meta.rdf_context = {'user_set': 'foaf:member'}
Group._meta.permission_classes = [(OwnerPermissions&ReadOnly)|InheritPermissions]
Group._meta.owner_field = 'user'
Group._meta.inherit_permissions = []

class LDPModelManager(models.Manager):
    def local(self):
        '''an alternative to all() which exlcudes external resources'''
        queryset = super(LDPModelManager, self).all()
        internal_ids = [x.pk for x in queryset if not Model.is_external(x)]
        return queryset.filter(pk__in=internal_ids)

class Model(models.Model):
    urlid = LDPUrlField(blank=True, null=True, unique=True, db_index=True)
    is_backlink = models.BooleanField(default=False, help_text='set automatically to indicate the Model is a backlink')
    allow_create_backlink = models.BooleanField(default=True,
                                                help_text='set to False to disable backlink creation after Model save')
    objects = LDPModelManager()

    class Meta:
        default_permissions = DEFAULT_DJANGOLDP_PERMISSIONS
        abstract = True
        depth = 0

    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)

    @classmethod
    def get_container_path(cls):
        '''returns the url path which is used to access actions on this model (e.g. /users/)'''
        path = getattr(cls._meta, 'container_path', getattr(cls.Meta, 'container_path', None))
        if path is None:
            path = "{}s".format(cls._meta.object_name.lower())

        return cls.__clean_path(path)

    def get_absolute_url(self):
        return Model.absolute_url(self)

    @classonlymethod
    def absolute_url(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase) or not instance_or_model.urlid:
            return '{}{}'.format(settings.SITE_URL, Model.resource(instance_or_model))
        else:
            return instance_or_model.urlid

    def get_container_id(self):
        return Model.container_id(self)

    @classonlymethod
    def resource(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase):
            return cls.container_id(instance_or_model)
        else:
            return cls.resource_id(instance_or_model)

    @classonlymethod
    def resource_id(cls, instance):
        r_id = "{}{}".format(cls.container_id(instance), getattr(instance, cls.slug_field(instance), ""))
        return cls.__clean_path(r_id)

    @classonlymethod
    def slug_field(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase):
            model = instance_or_model
        else:
            model = type(instance_or_model)
        
        # Use cached value if present
        if hasattr(model, "_slug_field"):
            return model._slug_field
        object_name = model.__name__.lower()
        view_name = '{}-detail'.format(object_name)

        try:
            slug_field = '/{}'.format(get_resolver().reverse_dict[view_name][0][0][1][0])
        except MultiValueDictKeyError:
            slug_field = getattr(model._meta, 'lookup_field', 'pk')
        if slug_field.startswith('/'):
            slug_field = slug_field[1:]
        
        model._slug_field = slug_field
        return slug_field

    @classonlymethod
    def container_id(cls, instance):
        if isinstance(instance, cls):
            path = instance.get_container_path()
        else:
            view_name = '{}-list'.format(instance._meta.object_name.lower())
            path = get_resolver().reverse(view_name)

        path = cls.__clean_path(path)

        return path

    @classonlymethod
    def resolve_id(cls, id):
        '''
        Resolves the id of a given path (e.g. /container/1/)
        Raises Resolver404 if the path cannot be found, ValidationError if the path is for a model base
        and an ObjectDoesNotExist exception if the resource does not exist
        '''
        id = cls.__clean_path(id)
        match = get_resolver().resolve(id)
        kwargs = match.kwargs
        view = match.func

        if match.url_name.endswith('-list') or len(match.kwargs.keys()) == 0:
            raise ValidationError('resolve_id received a path for a container or nested container')
        return view.initkwargs['model'].objects.get(**kwargs)

    @classonlymethod
    def resolve_parent(cls, path):
        split = path.strip('/').split('/')
        parent_path = "/".join(split[:-1])
        return Model.resolve_id(parent_path)

    @classonlymethod
    def resolve_container(cls, path):
        '''retruns the model container of passed URL path'''
        path = cls.__clean_path(path)
        view, args, kwargs = get_resolver().resolve(path)
        return view.initkwargs['model']

    @classonlymethod
    def resolve(cls, path):
        '''
        resolves the containing model and associated id in the path. If there is no id in the path returns None
        :param path: a URL path to check
        :return: the container model and resolved id in a tuple
        '''
        if path.startswith(settings.BASE_URL):
            path = path.replace(settings.BASE_URL, '')
        container = cls.resolve_container(path)
        try:
            resolve_id = cls.resolve_id(path)
        except:
            resolve_id = None
        return container, resolve_id

    @classonlymethod
    def __clean_path(cls, path):
        '''ensures path is Django-friendly'''
        if not path.startswith("/"):
            path = "/{}".format(path)
        if not path.endswith("/"):
            path = "{}/".format(path)
        return path

    @classonlymethod
    def get_or_create(cls, model, urlid, update=False, **field_tuples):
        '''
        gets an object with the passed urlid if it exists, creates it if not
        :param model: the model class which the object belongs to
        :param update: if set to True the object will be updated with the passed field_tuples
        :param field_tuples: kwargs for the model creation/updating
        :return: the object, fetched or created
        :raises Exception: if the object does not exist, but the data passed is invalid
        '''
        try:
            rval = model.objects.get(urlid=urlid)
            if update:
                for field in field_tuples.keys():
                    setattr(rval, field, field_tuples[field])
                rval.save()
            return rval
        except ObjectDoesNotExist:
            if model is get_user_model():
                field_tuples['username'] = str(uuid.uuid4())
            return model.objects.create(urlid=urlid, is_backlink=True, **field_tuples)

    @classonlymethod
    def get_or_create_external(cls, model, urlid, **kwargs):
        '''
        checks that the parameterised urlid is external and then returns the result of Model.get_or_create
        :raises ObjectDoesNotExist: if the urlid is not external and the object doesn't exist
        '''
        if not Model.is_external(urlid) and not model.objects.filter(urlid=urlid).exists():
            raise ObjectDoesNotExist
        return Model.get_or_create(model, urlid, **kwargs)

    @classonlymethod
    def get_subclass_with_rdf_type(cls, type):
        #TODO: deprecate
        '''returns Model subclass with Meta.rdf_type matching parameterised type, or None'''
        if type == 'foaf:user':
            return get_user_model()

        for subcls in Model.__subclasses__():
            if getattr(subcls._meta, "rdf_type", None) == type:
                return subcls

        return None

    @classmethod
    def is_external(cls, value):
        '''
        :param value: string urlid or an instance with urlid field
        :return: True if the urlid is external to the server, False otherwise
        '''
        try:
            if not value:
                return False
            if not isinstance(value, str):
                value = value.urlid

            # This expects all @ids to start with http which mlight not be universal. Maybe needs a fix.
            return value.startswith('http') and not value.startswith(settings.SITE_URL)
        except:
            return False

#TODO: this breaks the serializer, which probably assumes that traditional models don't have a urlid.
# models.Model.urlid = property(lambda self: '{}{}'.format(settings.SITE_URL, Model.resource(self)))

class LDPSource(Model):
    federation = models.CharField(max_length=255)

    class Meta(Model.Meta):
        rdf_type = 'sib:federatedContainer'
        ordering = ('federation',)
        container_path = 'sources'
        lookup_field = 'federation'

    def __str__(self):
        return "{}: {}".format(self.federation, self.urlid)


class Activity(Model):
    '''Models an ActivityStreams Activity'''
    local_id = LDPUrlField(help_text='/inbox or /outbox url (local - this server)')  # /inbox or /outbox full url
    external_id = LDPUrlField(null=True, help_text='the /inbox or /outbox url (from the sender or receiver)')
    payload = models.TextField()
    response_location = LDPUrlField(null=True, blank=True, help_text='Location saved activity can be found')
    response_code = models.CharField(null=True, blank=True, help_text='Response code sent by receiver', max_length=8)
    response_body = models.TextField(null=True)
    type = models.CharField(null=True, blank=True, help_text='the ActivityStreams type of the Activity',
                            max_length=64)
    is_finished = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    success = models.BooleanField(default=False, help_text='set to True when an Activity is successfully delivered')

    class Meta(Model.Meta):
        container_path = "activities"
        rdf_type = 'as:Activity'

    def to_activitystream(self):
        return json.loads(self.payload)

    def response_to_json(self):
        return self.to_activitystream()


# temporary database-side storage used for scheduled tasks in the ActivityQueue
class ScheduledActivity(Activity):
    failed_attempts = models.PositiveIntegerField(default=0,
                                                  help_text='a log of how many failed retries have been made sending the activity')

    def save(self, *args, **kwargs):
        self.is_finished = False
        super(ScheduledActivity, self).save(*args, **kwargs)


class Follower(Model):
    '''Models a subscription on a model. When the model is saved, an Update activity will be sent to the inbox'''
    object = models.URLField(help_text='the object being followed')
    inbox = models.URLField(help_text='the inbox recipient of updates')
    follower = models.URLField(help_text='(optional) the resource/actor following the object', blank=True)

    def __str__(self):
        return 'Inbox ' + str(self.inbox) + ' on ' + str(self.object)

class DynamicNestedField:
    '''
    Used to define a method as a nested_field.
    Usage:
        LDPUser.circles = lambda self: Circle.objects.filter(members__user=self)
        LDPUser.circles.field = DynamicNestedField(Circle, 'circles')
    '''
    related_query_name = None
    one_to_many = False
    many_to_many = True
    many_to_one = False
    one_to_one = False
    read_only = True
    name = ''
    def __init__(self, model:models.Model|None, remote_name:str, name:str='', remote:object|None=None) -> None:
        self.model = model
        self.name = name
        if remote:
            self.remote_field = remote
        else:
            self.remote_field = DynamicNestedField(None, '', remote_name, self)

@receiver([post_save])
def auto_urlid(sender, instance, **kwargs):
    if isinstance(instance, Model):
        changed = False
        if getattr(instance, Model.slug_field(instance), None) is None:
            setattr(instance, Model.slug_field(instance), instance.pk)
            changed = True
        if (not instance.urlid or 'None' in instance.urlid):
            instance.urlid = instance.get_absolute_url()
            changed = True
        if changed:
            instance.save()

@receiver(post_save)
def create_role_groups(sender, instance, created, **kwargs):
    if created:
        for name, params in getattr(instance._meta, 'permission_roles', {}).items():
            group = Group.objects.create(name=f'LDP_{instance._meta.model_name}_{name}_{instance.id}')
            setattr(instance, name, group)
            instance.save()
            if params.get('add_author'):
                assert hasattr(instance._meta, 'auto_author'), "add_author requires to also define auto_author"
                author = getattr(instance, instance._meta.auto_author)
                if author:
                    group.user_set.add(author)
            for permission in params.get('perms', []):
                assign_perm(f'{permission}_{instance._meta.model_name}', group, instance)


def invalidate_cache_if_has_entry(entry):
    from djangoldp.serializers import GLOBAL_SERIALIZER_CACHE

    if GLOBAL_SERIALIZER_CACHE.has(entry):
        GLOBAL_SERIALIZER_CACHE.invalidate(entry)

def invalidate_model_cache_if_has_entry(model):
    entry = getattr(model._meta, 'label', None)
    invalidate_cache_if_has_entry(entry)

@receiver([pre_save, pre_delete])
def invalidate_caches(sender, instance, **kwargs):
    invalidate_model_cache_if_has_entry(sender)

@receiver([m2m_changed])
def invalidate_caches_m2m(sender, instance, action, *args, **kwargs):
    invalidate_model_cache_if_has_entry(kwargs['model'])


from django.conf import settings
from django.db.models import Q
from rest_framework.filters import BaseFilterBackend

class OwnerFilterBackend(BaseFilterBackend):
    """Adds the objects owned by the user"""
    def filter_queryset(self, request, queryset, view):
        if request.user.is_superuser:
            return queryset
        if request.user.is_anonymous:
            return queryset.none()
        if getattr(view.model._meta, 'owner_field', None):
            return queryset.filter(**{view.model._meta.owner_field: request.user})
        if getattr(view.model._meta, 'owner_urlid_field', None):
            return queryset.filter(**{view.model._meta.owner_urlid_field: request.user.urlid})
        if getattr(view.model._meta, 'auto_author', None):
            return queryset.filter(**{view.model._meta.auto_author: request.user})
        return queryset

class PublicFilterBackend(BaseFilterBackend):
    """
    No filter applied.
    This class is useful for permission classes that don't filter objects, so that they can be chained with other
    """       
    def filter_queryset(self, request, queryset, view):
        public_field = queryset.model._meta.public_field
        return queryset.filter(**{public_field: True})

class NoFilterBackend(BaseFilterBackend):
    """
    No filter applied.
    This class is useful for permission classes that don't filter objects, so that they can be chained with other
    """       
    def filter_queryset(self, request, queryset, view):
        return queryset 

class LocalObjectFilterBackend(BaseFilterBackend):
    """
    Filter which removes external objects (federated backlinks) from the queryset
    For querysets which should only include local objects
    """
    def filter_queryset(self, request, queryset, view):
        return queryset.filter(urlid__startswith=settings.SITE_URL)


class LocalObjectOnContainerPathBackend(LocalObjectFilterBackend):
    """
    Override of LocalObjectFilterBackend which removes external objects when the view requested
    is the model container path
    """
    def filter_queryset(self, request, queryset, view):
        from djangoldp.models import Model

        if issubclass(view.model, Model) and request.path_info == view.model.get_container_path():
            return super(LocalObjectOnContainerPathBackend, self).filter_queryset(request, queryset, view)
        return queryset


class SearchByQueryParamFilterBackend(BaseFilterBackend):
    """
    Applies search fields in the request query params to the queryset
    """
    def filter_queryset(self, request, queryset, view):

        # the model fields on which to perform the search
        search_fields = request.GET.get('search-fields', None)
        # the terms to search the fields for
        search_terms = request.GET.get('search-terms', None)
        # the method of search to apply to the model fields
        search_method = request.GET.get('search-method', "basic")
        # union or intersection
        search_policy = request.GET.get('search-policy', 'union')

        # check if there is indeed a search requested
        if search_fields is None or search_terms is None:
            return queryset

        def _construct_search_query(search):
            '''Utility function pipes many Django Query objects'''
            search_query = []

            for idx, s in enumerate(search):
                if idx > 0:

                    # the user has indicated to make a union of all query results
                    if search_policy == "union":
                        search_query = search_query | Q(**s)

                    # the user has indicated to make an intersection of all query results
                    else:
                        search_query = search_query & Q(**s)

                    continue

                search_query = Q(**s)

            return search_query

        search_fields = search_fields.split(',')

        if search_method == "basic":
            search = []

            for s in search_fields:
                query = {}
                query["{}__contains".format(s)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        elif search_method == "ibasic":
            # NOTE: to use, see https://stackoverflow.com/questions/54071944/fielderror-unsupported-lookup-unaccent-for-charfield-or-join-on-the-field-not
            unaccent_extension = getattr(settings, 'SEARCH_UNACCENT_EXTENSION', False) and 'django.contrib.postgres' in settings.INSTALLED_APPS
            middle_term = '__unaccent' if unaccent_extension else ''

            search = []

            for s in search_fields:
                query = {}
                query["{}{}__icontains".format(s, middle_term)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        elif search_method == "exact":
            search = []

            for s in search_fields:
                query = {}
                query["{}__exact".format(s)] = search_terms
                search.append(query)

            queryset = queryset.filter(_construct_search_query(search))

        return queryset



import os
from django.apps import AppConfig

class DjangoldpConfig(AppConfig):
    name = 'djangoldp'

    def ready(self):
        self.auto_register_model_admin()
        self.start_activity_queue()
        
        # Patch guardian core to avoid prefetching permissions several times
        from guardian.core import ObjectPermissionChecker
        ObjectPermissionChecker._prefetch_cache_orig = ObjectPermissionChecker._prefetch_cache

        def _prefetch_cache(self):
            if hasattr(self.user, "_guardian_perms_cache"):
                self._obj_perms_cache = self.user._guardian_perms_cache
                return
            self._prefetch_cache_orig()

        ObjectPermissionChecker._prefetch_cache = _prefetch_cache

    def start_activity_queue(self):
        from djangoldp.activities.services import ActivityQueueService
        if os.environ.get('RUN_MAIN') is not None:
            ActivityQueueService.start()

    def auto_register_model_admin(self):
        '''
        Automatically registers Model subclasses in the admin panel (which have not already been added manually)
        '''
        from importlib import import_module

        from django.conf import settings
        from django.contrib import admin
        from djangoldp.admin import DjangoLDPAdmin
        from djangoldp.urls import get_all_non_abstract_subclasses
        from djangoldp.models import Model

        for package in settings.DJANGOLDP_PACKAGES:
            try:
                import_module('{}.admin'.format(package))
            except ModuleNotFoundError:
                pass

        for package in settings.DJANGOLDP_PACKAGES:
            try:
                import_module('{}.models'.format(package))
            except ModuleNotFoundError:
                pass

        for model in get_all_non_abstract_subclasses(Model):
            if not admin.site.is_registered(model):
                admin.site.register(model, DjangoLDPAdmin)



from importlib import import_module

from django.conf import settings
from django.contrib.auth.models import Group
from django.urls import path, re_path, include

from djangoldp.models import LDPSource, Model
from djangoldp.permissions import ReadOnly
from djangoldp.views import LDPSourceViewSet, WebFingerView, InboxView
from djangoldp.views import LDPViewSet

def __clean_path(path):
    '''ensures path is Django-friendly'''
    if path.startswith("/"):
        path = path[1:]
    if not path.endswith("/"):
        path = "{}/".format(path)
    return path


def get_all_non_abstract_subclasses(cls):
    '''
    returns a set of all subclasses for a given Python class (recursively calls cls.__subclasses__()). Ignores Abstract
    classes
    '''
    def valid_subclass(sc):
        '''returns True if the parameterised subclass is valid and should be returned'''
        return not getattr(sc._meta, 'abstract', False)

    return set(c for c in cls.__subclasses__() if valid_subclass(c)).union(
        [subclass for c in cls.__subclasses__() for subclass in get_all_non_abstract_subclasses(c) if valid_subclass(subclass)])

urlpatterns = [
    path('groups/', LDPViewSet.urls(model=Group)),
    re_path(r'^sources/(?P<federation>\w+)/', LDPSourceViewSet.urls(model=LDPSource, fields=['federation', 'urlid'],
                                                                    permission_classes=[ReadOnly], )),
    re_path(r'^\.well-known/webfinger/?$', WebFingerView.as_view()),
    path('inbox/', InboxView.as_view())
]

if settings.ENABLE_SWAGGER_DOCUMENTATION:
    from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView
    urlpatterns.extend([
        path("schema/", SpectacularAPIView.as_view(), name="schema"),
        path(
            "docs/",
            SpectacularSwaggerView.as_view(
                template_name="swagger-ui.html", url_name="schema"
            ),
            name="swagger-ui",
        )
    ])

for package in settings.DJANGOLDP_PACKAGES:
    try:
        import_module('{}.models'.format(package))
    except ModuleNotFoundError:
        pass
    try:
        urlpatterns.append(path('', include('{}.djangoldp_urls'.format(package))))
    except ModuleNotFoundError:
        pass

# append urls for all DjangoLDP Model subclasses
for model in get_all_non_abstract_subclasses(Model):
    # the path is the url for this model
    model_path = __clean_path(model.get_container_path())
    # urls_fct will be a method which generates urls for a ViewSet (defined in LDPViewSetGenerator)
    urls_fct = getattr(model, 'view_set', LDPViewSet).urls
    urlpatterns.append(path('' + model_path,
        urls_fct(model=model,
                 lookup_field=getattr(model._meta, 'lookup_field', 'pk'),
                 permission_classes=getattr(model._meta, 'permission_classes', []),
                 fields=getattr(model._meta, 'serializer_fields', []),
                 nested_fields=getattr(model._meta, 'nested_fields', [])
                 )))

# NOTE: this route will be ignored if a custom (subclass of Model) user model is used, or it is registered by a package
# Django matches the first url it finds for a given path
urlpatterns.append(re_path('users/', LDPViewSet.urls(model=settings.AUTH_USER_MODEL, permission_classes=[])))


from djangoldp.filters import BaseFilterBackend
from djangoldp.permissions import LDPBasePermission

class StartsWithAFilter(BaseFilterBackend):
    """Only objects whose title starts in A get through"""
    def filter_queryset(self, request, queryset, view):
        return queryset.filter(title__startswith='A')

class ReadOnlyStartsWithA(LDPBasePermission):
    """Only gives read-only access and only to objects which title starts with A"""
    filter_backend = StartsWithAFilter
    permissions = {'view', 'list'}
    def check_perms(self, obj):
        return getattr(obj, 'title', '').startswith('A')
    def has_object_permission(self, request, view, obj=None):
        return self.check_perms(obj)
    def get_permissions(self, user, model, obj=None):
        return self.permissions if self.check_perms(obj) else set()


class ContainsSpace(BaseFilterBackend):
    """Only objects whose title contains a space get through"""
    def filter_queryset(self, request, queryset, view):
        if request.user.username != 'toto':
            return queryset.none()
        return queryset.filter(title__contains=' ')

class Only2WordsForToto(LDPBasePermission):
    """Only gives access if the user's username is toto and only to objects whose title has two words (contains space)"""
    filter_backend = ContainsSpace
    def has_permission(self, request, view):
        return request.user.username == 'toto'
    def check_perms(self, obj):
        return ' ' in getattr(obj, 'title', '')
    def has_object_permission(self, request, view, obj=None):
        return self.check_perms(obj)
    def get_permissions(self, user, model, obj=None):
        return self.permissions if self.check_perms(obj) else set()


from django.contrib.auth import get_user_model

from rest_framework import status
from rest_framework.test import APIRequestFactory, APIClient, APITestCase
from djangoldp.tests.models import User, Circle, Project
from djangoldp.serializers import LDPSerializer
from djangoldp.related import get_prefetch_fields


class LDPViewSet(APITestCase):

    user_serializer_fields = ['@id', 'username', 'first_name', 'last_name', 'email', 'userprofile', 'conversation_set', 'projects']
    user_expected_fields = {'userprofile', 'conversation_set', 'projects', 'conversation_set__author_user', 'conversation_set__peer_user'}
    project_serializer_fields = ['@id', 'description', 'members']
    project_expected_fields = {'members', 'members__userprofile'}

    def setUpLoggedInUser(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion', first_name='John')
        self.client.force_authenticate(self.user)

    def _get_serializer(self, model, depth, fields):
        meta_args = {'model': model, 'depth': depth, 'fields': fields}
        meta_class = type('Meta', (), meta_args)
        return (type(LDPSerializer)('TestSerializer', (LDPSerializer,), {'Meta': meta_class}))()

    def test_get_prefetch_fields_user(self):
        model = User
        depth = 0
        serializer_fields = self.user_serializer_fields
        expected_fields = self.user_expected_fields
        serializer = self._get_serializer(model, depth, serializer_fields)
        result = get_prefetch_fields(model, serializer, depth)
        self.assertEqual(expected_fields, result)

    def test_get_prefetch_fields_circle(self):
        model = Circle
        depth = 0
        serializer_fields = ['@id', 'name', 'description', 'owner', 'members']
        expected_fields = {'owner', 'members', 'admins', 'space'}
        serializer = self._get_serializer(model, depth, serializer_fields)
        result = get_prefetch_fields(model, serializer, depth)
        self.assertEqual(expected_fields, result)

    def test_get_prefetch_fields_project(self):
        model = Project
        depth = 0
        serializer_fields = self.project_serializer_fields
        expected_fields = self.project_expected_fields
        serializer = self._get_serializer(model, depth, serializer_fields)
        result = get_prefetch_fields(model, serializer, depth)
        self.assertEqual(expected_fields, result)

    # TODO: dynamically generating serializer fields is necessary to retrieve many-to-many fields at depth > 0,
    #  but the _all_ default has issues detecting reverse many-to-many fields
    '''def test_get_prefetch_fields_depth_1(self):
        model = Project
        depth = 2
        serializer_fields = self.project_serializer_fields
        user_expected = set(['team__' + x for x in self.user_expected_fields])
        expected_fields = self.project_expected_fields.union(user_expected)
        serializer = self._get_serializer(model, depth, serializer_fields)
        result = get_prefetch_fields(model, serializer, depth)
        self.assertEqual(expected_fields, result)'''

    def test_get_shape_param(self):
        self.setUpLoggedInUser()
        circle = Circle.objects.create(name='test circle')

        # request id and name only
        fields_shape = '["@id", "name"]'

        response = self.client.get('/circles/', HTTP_ACCEPT_MODEL_FIELDS=fields_shape)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        response_data_keys = response.data['ldp:contains'][0].keys()
        self.assertIn(len(response_data_keys), (3,4))
        self.assertIn('@id', response_data_keys)
        self.assertIn('name', response_data_keys)
        self.assertIn('@type', response_data_keys)
        if len(response_data_keys)>3:
            self.assertIn('@context', response_data_keys)

    def test_search_fields_basic(self):
        self.setUpLoggedInUser()
        lowercase_circle = Circle.objects.create(name='test circle')
        uppercase_circle = Circle.objects.create(name='hello world', description='test')

        response = self.client.get('/circles/?search-fields=name&search-terms=test&search-method=basic')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        self.assertEqual(response.data['ldp:contains'][0]['name'], lowercase_circle.name)

        # test multiple search fields
        response = self.client.get('/circles/?search-fields=name,description&search-terms=test&search-method=basic')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['ldp:contains']), 2)
    
    def test_search_fields_ibasic(self):
        self.setUpLoggedInUser()
        lowercase_circle = Circle.objects.create(name='test circle')
        uppercase_circle = Circle.objects.create(name='TEST')

        response = self.client.get('/circles/?search-fields=name&search-terms=test&search-method=ibasic')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['ldp:contains']), 2)
    
    def test_search_fields_exact(self):
        self.setUpLoggedInUser()
        lowercase_circle = Circle.objects.create(name='test circle')
        uppercase_circle = Circle.objects.create(name='TEST')

        response = self.client.get('/circles/?search-fields=name&search-terms=test&search-method=exact')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        response = self.client.get('/circles/?search-fields=name&search-terms=test%20circle&search-method=exact')
        self.assertEqual(response.data['ldp:contains'][0]['name'], lowercase_circle.name)



from django.contrib.auth import get_user_model
from django.test import TestCase
from rest_framework.test import APIRequestFactory, APIClient
from rest_framework.utils import json

from djangoldp.models import Model
from djangoldp.tests.models import Invoice, LDPDummy, Resource, Post, Circle, Project, Space


class PostTestCase(TestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(self.user)

    def test_save_fk_graph_with_nested(self):
        post = {
            '@graph': [
                {
                    'https://cdn.startinblox.com/owl#title': "title",
                    'https://cdn.startinblox.com/owl#invoice': {
                        '@id': "_.123"
                    }
                },
                {
                    '@id': "_.123",
                    'https://cdn.startinblox.com/owl#title': "title 2"
                }
            ]
        }

        response = self.client.post('/batchs/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertNotIn('author', response.data)
        self.assertEquals(response.data['title'], "title")
        self.assertEquals(response.data['invoice']['title'], "title 2")

    def test_save_fk_graph_with_existing_nested(self):
        invoice = Invoice.objects.create(title="title 3")
        post = {
            '@graph': [
                {
                    'https://cdn.startinblox.com/owl#title': "title",
                    'https://cdn.startinblox.com/owl#invoice': {
                        '@id': "http://happy-dev.fr{}{}/".format(Model.container_id(invoice), invoice.id)
                    }
                }
            ]
        }

        response = self.client.post('/batchs/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertNotIn('author', response.data)
        self.assertEquals(response.data['title'], "title")
        self.assertEquals(response.data['invoice']['title'], "title 3")

    def test_post_should_accept_missing_field_id_nullable(self):
            body = [
                {
                    '@id': "./",
                    'https://cdn.startinblox.com/owl#content': "post update",
                }
            ]
            response = self.client.post('/posts/', data=json.dumps(body),
                                        content_type='application/ld+json')
            self.assertEqual(response.status_code, 201)
            self.assertIn('peer_user', response.data)

    def test_post_should_accept_empty_field_if_nullable(self):
        body = [
            {
                '@id': "./",
                'https://cdn.startinblox.com/owl#content': "post update",
                'https://cdn.startinblox.com/owl#peer_user': ""
            }
        ]
        response = self.client.post('/posts/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['peer_user'], None)

    def test_save_sub_object_in_new_object_with_reverse_1to1_relation(self):
        dummy = LDPDummy.objects.create(some="foo")

        body = [
            {
                '@id': "_:b216",
                'https://cdn.startinblox.com/owl#description': "user update",
                'https://cdn.startinblox.com/owl#ddummy': {
                    "@id": "http://happy-dev.fr{}{}/".format(Model.container_id(dummy), dummy.id)
                }
            },
            {
                '@id': './',
                "https://cdn.startinblox.com/owl#first_name": "Alexandre",
                "https://cdn.startinblox.com/owl#last_name": "Bourlier",
                "https://cdn.startinblox.com/owl#username": "alex",
                'https://cdn.startinblox.com/owl#userprofile': {'@id': "_:b216"}
            }
        ]
        response = self.client.post('/users/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertIn('userprofile', response.data)

    def test_embedded_context(self):
        body = {
            '@graph': [
                {
                    '@id': "./",
                    'content': "post update",
                    'peer_user': ""
                }
            ],
            '@context': {
                "@vocab": "https://cdn.startinblox.com/owl#",
            }
        }
        response = self.client.post('/posts/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

    def test_nested_container(self):
        resource = Resource.objects.create()
        body = {
            'https://cdn.startinblox.com/owl#title': "new job",
            'https://cdn.startinblox.com/owl#slug': "job1",
        }

        response = self.client.post('/resources/{}/joboffers/'.format(resource.pk),
                                    data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['resources']['ldp:contains'][0]['@id'], resource.urlid)
        self.assertEqual(response.data['title'], "new job")


    def test_nested_container_bis(self):
        invoice = Invoice.objects.create()
        body = {
            'https://cdn.startinblox.com/owl#title': "new batch",
        }

        response = self.client.post('/invoices/{}/batches/'.format(invoice.pk),
                                    data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['invoice']['@id'],
                         "http://happy-dev.fr/invoices/{}/".format(invoice.pk))
        self.assertEqual(response.data['title'], "new batch")

    def test_nested_container_federated(self):
        resource = Resource.objects.create()
        body = {
            'https://cdn.startinblox.com/owl#@id': "http://external.job/job/1",
        }

        response = self.client.post('/resources/{}/joboffers/'.format(resource.pk),
                                    data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['@id'], "http://external.job/job/1")
        self.assertIn('@type', response.data)
        response = self.client.get('/resources/{}/'.format(resource.pk))
        self.assertEqual(response.data['joboffers']['ldp:contains'][0]['@id'], "http://external.job/job/1")

    def test_embedded_context_2(self):
        body = {
            '@id': "./",
            'content': "post update",
            'peer_user': "",
            '@context': {
                "@vocab": "https://cdn.startinblox.com/owl#",
            }
        }

        response = self.client.post('/posts/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

    def test_auto_id(self):
        body = {
            '@id': "./",
            'content': "post update",
            'peer_user': "",
            '@context': {
                "@vocab": "https://cdn.startinblox.com/owl#",
            }
        }

        response = self.client.post('/posts/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        saved_post = Post.objects.get(pk=1)
        self.assertEqual(saved_post.urlid, "http://happy-dev.fr/posts/1/")

    def test_save_invalid_nested_user(self):
        body = {
            '@id': "./",
            'content': "post update",
            'peer_user': {'none': None},
            '@context': {
                "@vocab": "https://cdn.startinblox.com/owl#",
            }
        }

        response = self.client.post('/posts/', data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 400)

    def test_nested_container_user_federated(self):
        project = Project.objects.create()
        body = {
            'https://cdn.startinblox.com/owl#@id': "http://external.user/user/1/",
        }

        response = self.client.post('/projects/{}/members/'.format(project.pk),
                                    data=json.dumps(body),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['@id'], "http://external.user/user/1/")
        self.assertIn('@type', response.data)
        response = self.client.get('/projects/{}/'.format(project.pk))
        self.assertEqual(response.data['members']['ldp:contains'][0]['@id'], "http://external.user/user/1/")

    #  https://www.w3.org/TR/json-ld/#value-objects
    def test_post_field_with_value_object(self):
        post = {
            'https://cdn.startinblox.com/owl#title': {
                '@value': "title",
                '@language': "en"
            }
        }
        response = self.client.post('/invoices/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEquals(response.data['title'], "title")

    # from JSON-LD spec: "The value associated with the @value key MUST be either a string, a number, true, false or null"
    def test_save_field_with_invalid_value_object(self):
        invoice = Invoice.objects.create(title="title 3")
        post = {
            'https://cdn.startinblox.com/owl#invoice': {
                '@value': {'title': 'title',
                           '@id': "http://happy-dev.fr{}{}/".format(Model.container_id(invoice), invoice.id)}
            }
        }
        response = self.client.post('/batchs/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 400)

    # TODO: bug with PyLD: https://github.com/digitalbazaar/pyld/issues/142
    # from JSON-LD spec: "If the value associated with the @type key is @json, the value MAY be either an array or an object"
    '''
    def test_save_field_with_object_value_object(self):
        invoice = Invoice.objects.create(title="title 3")
        post = {
            'https://cdn.startinblox.com/owl#invoice': {
                '@value': {'title': 'title', '@id': "http://happy-dev.fr{}{}/".format(Model.container_id(invoice), invoice.id)},
                '@type': '@json'
            }
        }
        response = self.client.post('/batchs/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
    '''

    # the below test is necessary because of an obscure bug where the OneToOne field is successfully applied
    # during the life of the serializer (and response) but is not persisted in the database,
    # when it is posted onto the reverse relation
    def test_one_to_one_field_reverse_post(self):
        self.assertEqual(Circle.objects.count(), 0)
        self.assertEqual(Space.objects.count(), 0)

        body = {
            '@context': {'@vocab': "https://cdn.startinblox.com/owl#" },
            'space': {'name': "Etablissement"}
        }

        response = self.client.post('/circles/', data=json.dumps(body), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Circle.objects.count(), 1)
        self.assertEqual(Space.objects.count(), 1)

        circle = Circle.objects.all()[0]
        space = circle.space

        self.assertIsNotNone(space)
        self.assertIsNotNone(space.circle)



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission, Group
from django.conf import settings
from django.test import override_settings
from rest_framework.test import APIClient, APITestCase
from djangoldp.tests.models import JobOffer, LDPDummy, PermissionlessDummy, UserProfile, OwnedResource, \
    OwnedResourceNestedOwnership, OwnedResourceTwiceNestedOwnership

import json


class UserPermissionsTestCase(APITestCase):
    def setUp(self):
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com', password='glass onion')
        self.client = APIClient(enforce_csrf_checks=True)
        self.client.force_authenticate(user=self.user)
        self.job = JobOffer.objects.create(title="job", slug="slug1")


class TestUserPermissions(UserPermissionsTestCase):
    def setUpGroup(self):
        self.group = Group.objects.create(name='Test')
        view_perm = Permission.objects.get(codename='view_permissionlessdummy')
        self.group.permissions.add(view_perm)
        self.group.save()

    def _make_self_superuser(self):
        self.user.is_superuser = True
        self.user.save()

    # list - simple
    def test_get_for_authenticated_user(self):
        response = self.client.get('/job-offers/')
        self.assertEqual(response.status_code, 200)
        # test serialized permissions
        self.assertIn({'mode': {'@type': 'view'}}, response.data['permissions'])
        self.assertNotIn({'mode': {'@type': 'inherit'}}, response.data['permissions'])
        # self.assertNotIn({'mode': {'@type': 'delete'}}, response.data['permissions'])

    # TODO: list - I do not have permission from the model, but I do have permission via a Group I am assigned
    #  https://git.startinblox.com/djangoldp-packages/djangoldp/issues/291
    '''def test_group_list_access(self):
        self.setUpGroup()
        dummy = PermissionlessDummy.objects.create()

        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

        self.user.groups.add(self.group)
        self.user.save()
        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

    # repeat of the above test on nested field
    def test_group_list_access_nested_field(self):
        self.setUpGroup()
        parent = LDPDummy.objects.create()
        PermissionlessDummy.objects.create(parent=parent)

        response = self.client.get('/ldpdummys/{}/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['anons']['ldp:contains']), 0)

        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

        self.user.groups.add(self.group)
        self.user.save()
        response = self.client.get('/ldpdummys/{}/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['anons']['ldp:contains']), 1)

    # repeat of the test on a nested viewset
    def test_group_list_access_nested_viewset(self):
        self.setUpGroup()
        parent = LDPDummy.objects.create()
        PermissionlessDummy.objects.create(parent=parent)

        response = self.client.get('/ldpdummys/{}/anons/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

        self.user.groups.add(self.group)
        self.user.save()
        response = self.client.get('/ldpdummys/{}/anons/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

    # repeat for object-specific request
    def test_group_object_access(self):
        self.setUpGroup()
        dummy = PermissionlessDummy.objects.create()

        response = self.client.get('/permissionless-dummys/{}'.format(dummy))
        self.assertEqual(response.status_code, 404)

        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

        self.user.groups.add(self.group)
        self.user.save()
        response = self.client.get('/permissionless-dummys/{}/'.format(dummy))
        self.assertEqual(response.status_code, 200)
    
    # TODO: test for POST scenario
    # TODO: test for PUT scenario
    # TODO: test for DELETE scenario   
    '''

    @override_settings(SERIALIZE_OBJECT_EXCLUDE_PERMISSIONS=['inherit'])
    def test_get_1_for_authenticated_user(self):
        response = self.client.get('/job-offers/{}/'.format(self.job.slug))
        self.assertEqual(response.status_code, 200)
        self.assertIn({'mode': {'@type': 'view'}}, response.data['permissions'])
        self.assertNotIn({'mode': {'@type': 'inherit'}}, response.data['permissions'])

    def test_post_request_for_authenticated_user(self):
        post = {'https://cdn.startinblox.com/owl#title': "job_created", "https://cdn.startinblox.com/owl#slug": 'slug2'}
        response = self.client.post('/job-offers/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

    # denied because I don't have model permissions
    def test_post_request_denied_model_perms(self):
        data = {'https://cdn.startinblox.com/owl#some': 'title'}
        response = self.client.post('/permissionless-dummys/', data=json.dumps(data), content_type='application/ld+json')
        self.assertEqual(response.status_code, 403)

    def test_post_nested_view_authorized(self):
        data = { "https://cdn.startinblox.com/owl#title": "new skill", "https://cdn.startinblox.com/owl#obligatoire": "okay" }
        response = self.client.post('/job-offers/{}/skills/'.format(self.job.slug), data=json.dumps(data),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

    def test_post_nested_view_denied_model_perms(self):
        parent = LDPDummy.objects.create(some='parent')
        data = { "https://cdn.startinblox.com/owl#some": "title" }
        response = self.client.post('/ldpdummys/{}/anons/'.format(parent.pk), data=json.dumps(data),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 403)

    def test_put_request_for_authenticated_user(self):
        body = {'https://cdn.startinblox.com/owl#title':"job_updated"}
        response = self.client.put('/job-offers/{}/'.format(self.job.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

    def test_request_patch_for_authenticated_user(self):
        response = self.client.patch('/job-offers/' + str(self.job.slug) + "/",
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

    def test_put_request_denied_model_perms(self):
        dummy = PermissionlessDummy.objects.create(some='some', slug='slug')
        data = {'https://cdn.startinblox.com/owl#some': 'new'}
        response = self.client.put('/permissionless-dummys/{}/'.format(dummy.slug), data=json.dumps(data),
                                    content_type='application/ld+json')
        self.assertEqual(response.status_code, 404)

    def test_put_nested_view_denied_model_perms(self):
        parent = LDPDummy.objects.create(some='parent')
        child = PermissionlessDummy.objects.create(some='child', slug='child', parent=parent)
        data = {"https://cdn.startinblox.com/owl#some": "new"}
        response = self.client.put('/ldpdummys/{}/anons/{}/'.format(parent.pk, child.slug), data=json.dumps(data),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 404)

    #TODO: check how this could ever work
    # def test_patch_nested_container_attach_existing_resource_permission_denied(self):
    #     '''I am attempting to add a resource which I should not know exists'''
    #     parent = LDPDummy.objects.create(some='parent')
    #     dummy = PermissionlessDummy.objects.create(some='some', slug='slug')
    #     data = {
    #         'https://cdn.startinblox.com/owl#anons': [
    #             {'@id': '{}/permissionless-dummys/{}/'.format(settings.SITE_URL, dummy.slug), 'https://cdn.startinblox.com/owl#slug': dummy.slug}
    #         ]
    #     }
    #     response = self.client.patch('/ldpdummys/{}/'.format(parent.pk), data=json.dumps(data), content_type='application/ld+json')
    #     self.assertEqual(response.status_code, 404)

    # variations on previous tests with an extra level of depth
    # TODO
    def test_post_nested_container_twice_nested_permission_denied(self):
        pass

    # TODO
    def test_put_nested_container_twice_nested_permission_denied(self):
        pass

    # TODO: repeat of the above where it is authorized because I have permission through my Group
    #  https://git.startinblox.com/djangoldp-packages/djangoldp/issues/291

    def test_put_request_change_urlid_rejected(self):
        self.assertEqual(JobOffer.objects.count(), 1)
        body = {'@id': "ishouldnotbeabletochangethis"}
        response = self.client.put('/job-offers/{}/'.format(self.job.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        # TODO: this is failing quietly
        #  https://git.happy-dev.fr/startinblox/solid-spec/issues/14
        self.assertEqual(response.status_code, 200)
        self.assertEqual(JobOffer.objects.count(), 1)
        self.assertFalse(JobOffer.objects.filter(urlid=body['@id']).exists())

    def test_put_request_change_pk_rejected(self):
        self.assertEqual(JobOffer.objects.count(), 1)
        body = {'https://cdn.startinblox.com/owl#pk': 2}
        response = self.client.put('/job-offers/{}/'.format(self.job.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        # TODO: this is failing quietly
        #  https://git.happy-dev.fr/startinblox/solid-spec/issues/14
        self.assertEqual(response.status_code, 200)
        self.assertEqual(JobOffer.objects.count(), 1)
        self.assertFalse(JobOffer.objects.filter(pk=body['https://cdn.startinblox.com/owl#pk']).exists())

    # tests that I receive a list of objects for which I am owner, filtering those for which I am not
    def test_list_owned_resources(self):
        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.get('/ownedresources/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        self.assertEqual(response.data['ldp:contains'][0]['@id'], my_resource.urlid)

    # I do not have model permissions as an authenticated user, but I am the resources' owner
    def test_get_owned_resource(self):
        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.get('/ownedresources/{}/'.format(my_resource.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['@id'], my_resource.urlid)
        self.assertIn({'mode': {'@type': 'delete'}}, response.data['permissions'])

        # I have permission to view this resource
        response = self.client.patch('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 404)

    def test_patch_owned_resource(self):
        my_profile = UserProfile.objects.create(user=self.user, slug=self.user.username, description='about me')
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_profile = UserProfile.objects.create(user=another_user, slug=another_user.username, description='about')

        response = self.client.patch('/userprofiles/{}/'.format(my_profile.slug))
        self.assertEqual(response.status_code, 200)

        response = self.client.patch('/userprofiles/{}/'.format(their_profile.slug))
        self.assertEqual(response.status_code, 403)

    def test_delete_owned_resource(self):
        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.delete('/ownedresources/{}/'.format(my_resource.pk))
        self.assertEqual(response.status_code, 204)

        response = self.client.delete('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 404)

    # test superuser permissions (configured on model)
    def test_list_superuser_perms(self):
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.get('/ownedresources/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        # now I'm superuser, I have the permissions
        self._make_self_superuser()

        response = self.client.get('/ownedresources/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

    def test_get_superuser_perms(self):
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.patch('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 404)

        self._make_self_superuser()

        response = self.client.patch('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 200)

    def test_put_superuser_perms(self):
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_profile = UserProfile.objects.create(user=another_user, slug=another_user.username, description='about')

        response = self.client.patch('/userprofiles/{}/'.format(their_profile.slug))
        # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/336
        self.assertEqual(response.status_code, 403)

        self._make_self_superuser()

        response = self.client.patch('/userprofiles/{}/'.format(their_profile.slug))
        self.assertEqual(response.status_code, 200)

    def test_delete_superuser_perms(self):
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        response = self.client.delete('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 404)

        self._make_self_superuser()

        response = self.client.delete('/ownedresources/{}/'.format(their_resource.pk))
        self.assertEqual(response.status_code, 204)

    # I have model (or object?) permissions. Attempt to make myself owner and thus upgrade my permissions
    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/356/
    '''
    def test_hack_model_perms_privilege_escalation(self):
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        resource = OwnedResourceVariant.objects.create(description='another test', user=another_user)

        # authenticated has 'change' permission but only owner's have 'control' permission, meaning that I should
        # not be able to change my privilege level
        body = {
            'https://cdn.startinblox.com/owl#user': {'@id': self.user.urlid}
        }
        response = self.client.put('/ownedresourcevariants/{}/'.format(resource.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        resource = OwnedResourceVariant.objects.get(pk=resource.pk)
        self.assertNotEqual(resource.user, self.user)
    '''


class TestOwnerFieldUserPermissions(UserPermissionsTestCase):
    restore_meta = None

    def setUpTempOwnerFieldForModel(self, model, new_owner_field):
        # store the old meta information for tearDown to cleanup after the test
        if self.restore_meta is None:
            self.restore_meta = []
        
        self.restore_meta.append({
            "model": model,
            "owner_field": model._meta.owner_field
        })

        # replace the owner_field attribute for the test to run
        model._meta.owner_field = new_owner_field

    def tearDown(self):
        # restore any previously changed owner_field attributes in the test
        if self.restore_meta is not None:
            for idx, model in enumerate(self.restore_meta):
                model = self.restore_meta[idx]["model"]
                model._meta.owner_field = self.restore_meta[idx]["owner_field"]
            self.restore_meta = None

    def test_list_owned_resources_nested(self):
        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        my_second_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        my_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_resource)
        my_second_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_second_resource)
        their_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=their_resource)

        response = self.client.get('/ownedresourcenestedownerships/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 2)
        ids = [r['@id'] for r in response.data['ldp:contains']]
        self.assertIn(my_nested.urlid, ids)
        self.assertIn(my_second_nested.urlid, ids)
        self.assertNotIn(their_nested.urlid, ids)
    
    def test_list_owned_resources_nested_variation_urlid(self):
        owner_field = OwnedResourceNestedOwnership._meta.owner_field
        OwnedResourceNestedOwnership._meta.owner_field = None
        OwnedResourceNestedOwnership._meta.owner_urlid_field = owner_field + "__urlid"

        self.test_list_owned_resources_nested()
        OwnedResourceNestedOwnership._meta.owner_urlid_field = None
        OwnedResourceNestedOwnership._meta.owner_field = owner_field

    
    def test_list_owned_resources_nested_variation_twice_nested(self):
        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        my_second_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        my_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_resource)
        my_second_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_second_resource)
        their_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=their_resource)

        my_twice_nested = OwnedResourceTwiceNestedOwnership.objects.create(description="test", parent=my_nested)
        their_twice_nested = OwnedResourceTwiceNestedOwnership.objects.create(description="test", parent=their_nested)

        response = self.client.get('/ownedresourcetwicenestedownerships/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        ids = [r['@id'] for r in response.data['ldp:contains']]
        self.assertIn(my_twice_nested.urlid, ids)
        self.assertNotIn(their_twice_nested.urlid, ids)

    def test_list_owned_resources_nested_does_not_exist(self):
        self.setUpTempOwnerFieldForModel(OwnedResourceNestedOwnership, "parent__doesnotexist")

        my_resource = OwnedResource.objects.create(description='test', user=self.user)
        my_second_resource = OwnedResource.objects.create(description='test', user=self.user)
        another_user = get_user_model().objects.create_user(username='test', email='test@test.com', password='test')
        their_resource = OwnedResource.objects.create(description='another test', user=another_user)

        my_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_resource)
        my_second_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=my_second_resource)
        their_nested = OwnedResourceNestedOwnership.objects.create(description="test", parent=their_resource)

        self.assertRaises(ValueError, self.client.get, '/ownedresourcenestedownerships/')



import click
import sys
import yaml
import subprocess
from pkg_resources import resource_filename
from pathlib import Path
from django.core import management
from django.core.management.base import CommandError
from . import __version__


# click entrypoint
@click.group()
@click.version_option(__version__)
def main():

    """DjangoLDP CLI"""


@main.command()
@click.argument('name', nargs=1, required=False)
def initserver(name):

    """Start a DjangoLDP server."""

    try:
        # use directly pwd
        directory = Path.cwd()

        if name:
            # create a directory from project name in pwd
            directory = Path.cwd() / name
            directory.mkdir(parents=False, exist_ok=False)

        # get the template path
        template = resource_filename(__name__, 'conf/server_template')

        # wrap the default django-admin startproject command
        # this call import django settings and configure it
        # see: https://docs.djangoproject.com/fr/2.2/topics/settings/#calling-django-setup-is-required-for-standalone-django-usage
        # see: https://github.com/django/django/blob/stable/2.2.x/django/core/management/templates.py#L108
        # fix: in 2.2 gabarit files options has been renamed: https://github.com/django/django/blob/stable/2.2.x/django/core/management/templates.py#L53
        management.call_command('startproject', name, directory, template=template, files=['settings.yml'])

    except FileExistsError:
        click.echo(f'Error: the folder {directory} already exists')
        sys.exit(1)

    except CommandError as e:
        click.echo(f'Error: {e}')
        directory.rmdir()
        sys.exit(1)

@main.command()
@click.argument('name', nargs=1)
def startpackage(name):

    """Start a DjangoLDP package."""

    try:
        # set directory
        directory = Path.cwd() / name

        # get the template path
        template = resource_filename(__name__, 'conf/package_template')

        # create dir
        directory.mkdir(parents=False, exist_ok=False)

        # wrap the default startapp command
        management.call_command('startapp', name, directory, template=template)

    except FileExistsError:
        click.echo(f'Error: the folder {directory} already exists')
        sys.exit(1)

    except CommandError as e:
        click.echo(f'Error: {e}')
        sys.exit(1)

@main.command()
def install():

    """Install project dependencies."""

    try:
        # load dependencies from config file
        path = Path.cwd() / 'settings.yml'
        with open(path, 'r') as f:
            dependencies = yaml.safe_load(f).get('dependencies', [])

        # install them by calling pip command
        cmd = [sys.executable, "-m", "pip", "install", "--upgrade"]
        try:
            cmd.extend(dependencies)
            subprocess.run(cmd).check_returncode()
            click.echo('Installation done!')
        except TypeError:
            click.echo('No dependency to install')

    except FileNotFoundError:
        click.echo('Config error: no settings.yml file in this directory')
        sys.exit(1)

    except subprocess.CalledProcessError as e:
        click.echo(f'Installation error: {e}')
        sys.exit(1)


@main.command()
@click.option('--with-admin', 'admin', help='Create an administrator user with email.')
@click.option('--email', help='Provide an email for administrator.')
@click.option('--with-dummy-admin', 'dummy_admin', is_flag=True, help='Create a default "admin" user.')
def configure(admin, dummy_admin, email):

    """Configure the project."""

    try:
        # shortcut to the djangoldp.management command
        path = str(Path.cwd() / 'manage.py')
        cmd = [sys.executable, path, 'configure']
        if admin:
            if not email:
                click.echo('Error: missing email for admin user')
                return
            cmd.extend(['--with-admin', admin, '--email', email])
        elif dummy_admin:
            cmd.append('--with-dummy-admin')
        subprocess.run(cmd).check_returncode()

        click.echo('Configuration done!')

    except subprocess.CalledProcessError as e:
        click.echo(f'Configuration error: {e}')
        sys.exit(1)


@main.command()
def runserver():

    """Run the Django embeded webserver."""

    try:
        # shortcut to the djangoldp.management command
        path = str(Path.cwd() / 'manage.py')
        cmd = [sys.executable, path, 'runserver', '0.0.0.0:8000']
        subprocess.run(cmd).check_returncode()

    except subprocess.CalledProcessError as e:
        click.echo(f'Execution error: {e}')
        sys.exit(1)



from django.db.models import options

__version__ = '3.1.4'

options.DEFAULT_NAMES += (
    'lookup_field', 'rdf_type', 'rdf_context', 'auto_author', 'owner_field', 'owner_urlid_field',
    'view_set', 'container_path', 'permission_classes', 'serializer_fields', 'serializer_fields_exclude', 'empty_containers',
    'nested_fields', 'depth', 'permission_roles', 'inherit_permissions', 'public_field')



# Generated by Django 4.2.3 on 2023-10-17 19:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0017_alter_activity_urlid_alter_follower_urlid_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='payload',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='activity',
            name='response_body',
            field=models.TextField(null=True),
        ),
    ]



from django.urls import path
from djangoldp.tests.models import Message, Conversation, Dummy, PermissionlessDummy, Task, DateModel, LDPDummy
from djangoldp.permissions import ACLPermissions
from djangoldp.views import LDPViewSet

urlpatterns = [
    path('messages/', LDPViewSet.urls(model=Message, fields=["@id", "text", "conversation"], nested_fields=['conversation'])),
    path('tasks/', LDPViewSet.urls(model=Task)),
    path('conversations/', LDPViewSet.urls(model=Conversation, nested_fields=["message_set", "observers"])),
    path('dummys/', LDPViewSet.urls(model=Dummy, lookup_field='slug',)),
    path('permissionless-dummys/', LDPViewSet.urls(model=PermissionlessDummy, lookup_field='slug', permission_classes=[ACLPermissions])),
]




from django.conf import settings
from guardian.utils import get_anonymous_user


# convenience function returns True if user is anonymous
def is_anonymous_user(user):
    anonymous_username = getattr(settings, 'ANONYMOUS_USER_NAME', None)
    return user.is_anonymous or user.username == anonymous_username


# convenience function returns True if user is authenticated
def is_authenticated_user(user):
    anonymous_username = getattr(settings, 'ANONYMOUS_USER_NAME', None)
    return user.is_authenticated and user.username != anonymous_username



import json
import uuid
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from rest_framework.test import APIClient, APITestCase
from guardian.shortcuts import assign_perm

from .models import PermissionlessDummy, Dummy, LDPDummy


class TestsGuardian(APITestCase):

    def setUp(self):
        self.client = APIClient(enforce_csrf_checks=True)

    def setUpLoggedInUser(self):
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.group = Group.objects.create(name='Test')
        self.user.groups.add(self.group)
        self.user.save()
        self.client.force_authenticate(user=self.user)

    def _get_dummy_with_perms(self, perms=None, parent=None, group=False):
        if perms is None:
            perms = []
        dummy = PermissionlessDummy.objects.create(some='test', slug=uuid.uuid4(), parent=parent)
        model_name = PermissionlessDummy._meta.model_name

        for perm in perms:
            perm = perm + '_' + model_name
            if group:
                #assigns container-level and object-level perms
                assign_perm('tests.'+perm, self.group)
                assign_perm(perm, self.group, dummy)
            else:
                assign_perm('tests.'+perm, self.user)
                assign_perm(perm, self.user, dummy)

        return dummy

    # optional setup for testing PermissionlessDummy model with parameterised perms
    def setUpGuardianDummyWithPerms(self, perms=None, parent=None, group=False):
        self.dummy = self._get_dummy_with_perms(perms, parent, group)

    # auxiliary function converts permission format for test
    def _unpack_permissions(self, perms_from_response):
        return [p['mode']['@type'] for p in perms_from_response]

    # test that dummy with no permissions set returns no results
    def test_get_dummy_no_permissions(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms()
        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertEqual(response.status_code, 404)

    # test with anonymous user
    def test_get_dummy_anonymous_user(self):
        self.setUpGuardianDummyWithPerms()
        response = self.client.get('/permissionless-dummys/')
        # I have no object permissions - I should receive a 403
        self.assertEqual(response.status_code, 403)

    def test_list_dummy_exception(self):
        self.setUpLoggedInUser()
        # I have permission on a permissionless dummy, but not in general
        dummy_a = self._get_dummy_with_perms()
        dummy_b = self._get_dummy_with_perms(['view'])
        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        containees = [d['@id'] for d in response.data['ldp:contains']]
        self.assertNotIn(dummy_a.urlid, containees)
        self.assertIn(dummy_b.urlid, containees)

    def test_list_dummy_group_exception(self):
        self.setUpLoggedInUser()
        dummy_a = self._get_dummy_with_perms()
        dummy_b = self._get_dummy_with_perms(['view'], group=True)
        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        containees = [d['@id'] for d in response.data['ldp:contains']]
        self.assertNotIn(dummy_a.urlid, containees)
        self.assertIn(dummy_b.urlid, containees)

    def test_list_dummy_exception_nested_view(self):
        self.setUpLoggedInUser()
        parent = LDPDummy.objects.create(some="test")
        # two dummies, one I have permission to view and one I don't
        dummy_a = self._get_dummy_with_perms(parent=parent)
        dummy_b = self._get_dummy_with_perms(['view'], parent)
        response = self.client.get('/ldpdummys/{}/anons/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

    def test_list_dummy_exception_nested_serializer(self):
        self.setUpLoggedInUser()
        parent = LDPDummy.objects.create(some="test")
        # two dummies, one I have permission to view and one I don't
        dummy_a = self._get_dummy_with_perms(parent=parent)
        dummy_b = self._get_dummy_with_perms(['view'], parent)
        response = self.client.get('/ldpdummys/{}/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['anons']['ldp:contains']), 1)

    def test_get_dummy_permission_granted(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view'])
        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertEqual(response.status_code, 200)

    def test_get_dummy_group_permission_granted(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view'], group=True)
        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertEqual(response.status_code, 200)

    def test_get_dummy_permission_rejected(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view'])
        dummy_without = PermissionlessDummy.objects.create(some='test2', slug='test2')
        response = self.client.get('/permissionless-dummys/{}/'.format(dummy_without.slug))
        self.assertEqual(response.status_code, 404)

    def test_patch_dummy_permission_granted(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view', 'change'])
        body = {'some': "some_new"}
        response = self.client.patch('/permissionless-dummys/{}/'.format(self.dummy.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

    def test_patch_dummy_permission_rejected(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['change'])
        dummy_without = PermissionlessDummy.objects.create(some='test2', slug='test2')
        body = {'some': "some_new"}
        response = self.client.patch('/permissionless-dummys/{}/'.format(dummy_without.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 404)

    # TODO: PUT container of many objects approved on specific resource for which I do not have _model_ permissions

    # test that custom permissions are returned on a model
    def test_custom_permissions(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['custom_permission', 'view'])

        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertIn('custom_permission', self._unpack_permissions(response.data['permissions']))

    # test that duplicate permissions aren't returned
    def test_no_duplicate_permissions(self):
        self.setUpLoggedInUser()
        dummy = Dummy.objects.create(some='test', slug='test')
        model_name = Dummy._meta.model_name

        assign_perm('view_' + model_name, self.user, dummy)

        response = self.client.get('/dummys/{}/'.format(dummy.slug))
        self.assertEqual(response.status_code, 200)
        perms = self._unpack_permissions(response.data['permissions'])
        self.assertIn('view', perms)
        view_perms = [perm for perm in perms if perm == 'view']
        self.assertEqual(len(view_perms), 1)

    # TODO: attempting to migrate my object permissions by changing FK reference
